﻿import numpy as np
import torch
import datetime
from   sklearn.decomposition import PCA
from   sklearn import linear_model
import matplotlib.pyplot as plt                   # постороение графиков


class Embedding():
    """Загрузчик векторов и словаря слов из PyTorch формата *.pt """
    def __init__(self, wordID=None, words=None, start=5):      
        self.vectors  = None     # PyTorch - матрица векторов
        self.wordID   = wordID   # словарь слов  {the: {'id': 6, 'pm': 43976.69},...}
        self.words    = words    # список слов   ['the', ...]    
        self.startID  = start    # номер начала осмысленных слов (после пропуска служебных)
        self.metric   = 0        # расстояние  0: 1-cos(v,u), 1: евклидово расстояние sqrt((u-v)^2)

    def word(self, idx):
        """Получить слово по его номеру"""
        return self.words[idx]

    def idx(self, word):
        """Получить индекс слова. Если его нет - это 1: <UNK>"""
        if word in self.wordID:
            return self.wordID[word]["id"]
        return 1                             # <UNK>    
    
    def vec(self, word):
        """Получить вектор по слову"""
        return self.vectors[self.idx(word)]    

    def __call__(self, word):
        """Получить вектор по слову (скобочный оператор)"""
        return self.vectors[self.wordID[word]["id"]]        

    def dist(self, u, v):
        """ Метрика ближайшего расстояния между векторами """
        if self.metric == 0: 
            return self.dist_cos(u, v)
        return self.dist_len(u, v)

    def dist_cos(self, u, v):
        """ Метрика ближайшего расстояния между векторами """
        return  1 - torch.dot(u,v)/max(torch.sqrt(torch.dot(u,u)*torch.dot(v,v)), 1.e-8)

    def dist_len(self, u, v):
        """ Метрика ближайшего евклидового расстояния (корень из суммы квадратов) между векторами """
        return torch.norm(u-v)

    
    def save(self, fname, info):        
        """Сохранить в файле"""
        state = {'info'     :   info,
                 'date'     :   datetime.datetime.now(),
                 'words2id' :   self.wordID,
                 'vectors'  :   self.vectors
                }  
        torch.save(state, fname)  
        
    def load(self, fname):
        """Загрузить из файла"""
        state = torch.load(fname)  
        
        print("info:", state['info'])        
        print("date:", state['date'])
        
        self.wordID = state['words2id']
        self.vectors  = state['vectors']
        print("vectors:", self.vectors.shape)
        
        self.words  = []
        i = 0
        for k,v in self.wordID.items():
            if i <= 10:  print("%-10s"%(k),v, self.vectors[i][:3]); 
            i += 1
            self.words.append(k)                        
                
    def load_Glove(self, fname, max_words=1000):
        """Загрузить вектора из текстового файла Glove"""
        vec = {}
        self.words  = []
        with open(fname, "r", encoding='utf-8') as file:
            NUM_SPECIAL = 5
            self.words  = ["<NUL>",         "<UNK>",         "<BOS>",         "<SEP>",         "<CVR>"]
            self.wordID = {"<NUL>":{'id':0},"<UNK>":{'id':1},"<BOS>":{'id':2},"<SEP>":{'id':3},"<CVR>":{'id':4}} 
            i = 0
            for line in file:                
                if i < max_words:                                    
                    values = line.split()   # values[0] - слово; values[1:] - компоненты
                    self.words.append(values[0])
                    self.wordID[values[0]] = { "id": i+NUM_SPECIAL, "pm": int(100*max_words/(i+1)) }
                
                    vec = np.asarray(values[1:], dtype='float32')
                    if i == 0:
                        self.vectors = torch.zeros(max_words+NUM_SPECIAL, vec.shape[0])
                    self.vectors[i+NUM_SPECIAL].copy_(torch.from_numpy(vec))
                i += 1
            print("num_lines:", i)
        
                
    def aver_vector(self, words):
        """ Средний вектор списка слов words """
        aver = torch.zeros( self.vectors.shape[1] )
        for w in words:
            id =  self.wordID[w]["id"]
            aver += self.vectors[id, :]
        return aver / len(words)    
    
    def claster_size(self, words):
        """ Вычисляет среднюю компактность кластера слов words """        
        center = self.aver_vector(words)
        s = 0
        for w in words:
            v =  self.vectors[self.wordID[w]["id"], :]
            s += self.dist(center, v)
        return s/len(words)    
                    
    def get_nearest(self, vec, reverse=False):
        """ Получаем список всех слов, отсортированный по убыванию от точки vec """                    
        lst = []            
        for i in range(self.startID, self.vectors.shape[0]):
            v = self.vectors[i, :]
            lst.append( (i, self.dist(vec, v) ) )
        return sorted(lst, key=lambda kv: kv[1], reverse=reverse) 

    def print_nearest(self, lst, count=10, only_words=False, pm_min=0, no_word=""): 
        """ Выводим в форматном виде список """
        res = []
        for i in lst[:count]:
            w = self.words[i[0]]
            if w == no_word:
                continue
            if self.wordID[w]["pm"] > pm_min:
                if only_words:
                    res += ["%s" % ( w )]
                else:
                    res += ["%s (%.2f)" % ( w, i[1] )]
        if only_words:
            print( sorted(res) )
        else:
            print(res)
        
    def nearest_to_word(self, word, count=10, pm_min=0, reverse=False):
        """Вывести список слов, ближайших к слову word"""
        id = self.wordID[word]["id"]
        print("*", word, end=": ")
        self.print_nearest(self.get_nearest(self.vectors[id, :], reverse=reverse), count, False, pm_min, word)                
    
    def nearest_to_list(self, words, count=10, only_words=False, pm_min=0, reverse=False):
        """Вывести список слов, ближайших среднему слову из списка words"""
        aver = self.aver_vector(words)
        self.print_nearest( self.get_nearest(aver, reverse=reverse), count, only_words, pm_min)    
                
    def nearest_to_vec(self, vec, count=10, only_words=False, pm_min=0, reverse=False):
        """Вывести список слов, ближайших среднему слову из списка words"""
        self.print_nearest( self.get_nearest(vec, reverse=reverse), count, only_words, pm_min)            
        
    def  bin_classifier_BOW(self,  doc, s1, s2, stop =  {'the', 'a', 'an', ',', '!', '.'}):                    
        s1 = set( s1.split() ) - stop                     # правильное предложение
        s2 = set( s2.split() ) - stop                     # неправильное предложение                
        ss = set( doc.split()) - stop         
        return int(  len( s1 & ss ) >  len(s2 & ss ) )    # число пересечений            
        
    def sent_to_vec(self, s):
        s = [w for w in s.split() if w in self.wordID]
        return self.aver_vector(s)
                    
    def pca(self, docs, E_DIM):
        """ Построение векторизации методом контекстных вероятностей и снижения размерности """
        V_DIM = len(self.wordID)
        p12 = np.zeros((V_DIM, V_DIM))                      # совместные вероятности                         
        p1  = np.zeros((V_DIM,))                            # частоты слов        
        
        print("Calc PMI... len(docs):", len(docs), "  V_DIM:",V_DIM)
        for d in docs:                                      # по документам
            for s in d:                                     # по предложениям документа    
                ws = set(s.split())                         # "словарь" предложения        
                for w1 in ws:                                                           
                    if w1 in self.wordID:
                        p1[self.idx(w1)] += 1
                        for w2 in ws:                                
                            if w2 in self. wordID:
                                p12[ self.idx(w1), self.idx(w2) ] += 1   
                                    
        p12 /= p12.sum()
        p1  /= p1.sum()
        p1[:self.startID] = 1
        
        p12 = p12 / ( p1.reshape((V_DIM,1)) @ p1.reshape((1, V_DIM)) )        
        p12[p12 <= 0] = 1          
        p12 = np.log(p12) 
        p12[p12 <= 0] = 0
        p12 = p12 - p12.mean(axis=0)

        print("Calc PCA... max(p12):", f"{p12.max():.3f}", "  dtype:", p12.dtype)

        pca = PCA()
        res = pca.fit_transform(p12) 
        
        print("Normalize...")
        
        res = res[:, : E_DIM]        
        res = (res-res.mean(axis=0)) / (res.std(axis=0) * np.sqrt(E_DIM) )
        
        self.vectors = torch.tensor(res, dtype=torch.float )
        if "<CVR>" in self.wordID:
            self.vectors[:5,:].zero_()
        
        return pca.singular_values_
        
    def plot_hist_length(self, num = 1000, skip = 5):
        """
        Построить гистограмму распределения длин векторов, разбив их на редкие слова,
        частые и все.
        """
        len_common, len_seldom, len_all = [],[],[]
        
        vec = self.vectors        
        for i,w in enumerate(self.wordID):        
            if i < skip: continue
            if i < num:          len_common.append(vec[i].norm()) 
            if i > len(vec)-num: len_seldom.append(vec[i].norm()) 
            len_all.append(vec[i].norm()) 
            
        plt.figure(figsize=(16,3)) 
        
        plt.subplot(1,3,1); 
        plt.title (f'частые слова id < {num}'); plt.ylabel('density'); plt.xlabel('vector length')
        plt.hist(len_common, 30, density=True); 
        
        plt.subplot(1,3,2); 
        plt.title (f'редкие слова id > {len(vec)-num}'); plt.ylabel('density'); plt.xlabel('vector length')
        plt.hist(len_seldom, 30, density=True); 
   
        plt.subplot(1,3,3); 
        plt.title (f'все слова'); plt.ylabel('density'); plt.xlabel('vector length')
        plt.hist(len_all, 30, density=True); 
   
        plt.show()        

    def get_cos_dis(self,lst):   
        """ для пар слов из списка lst получаем 1-cos, и dist_len расстояния """
        dis1, dis2 = [],[]
        vec = self.vectors   
        for i in range(len(lst)):
            for j in range(i) :
                v1, v2 = vec[lst[i]], vec[lst[j]]                  
                dis1.append(self.dist_len(v1, v2).item());  
                dis2.append(self.dist_cos(v1, v2).item());   
        return dis1, dis2
        
    def plot_cos_dis(self, num = 1000, skip = 5):   
        """ Зависимость расстояния от 1-косинуса для первых num и последних num слов"""
        lst1, lst2 = [],[]
        vec = self.vectors   
        for i,w in enumerate(self.wordID):        
            if i < skip: continue
            if i < num:            lst1.append(i) 
            if i > len(vec )-num:  lst2.append(i)  
            
        plt.figure(figsize=(16,5))          
        for i, ls in enumerate(zip([lst1, lst2, lst1+lst2], [f"частые id < {num}", f"редкие id > {len(vec)-num}","их смесь"])):
            plt.subplot(1, 3, i+1)               
            dis1, dis2 = self.get_cos_dis(ls[0])
            
            plt.title(ls[1])
            plt.ylabel('cos'); plt.xlabel('dis')
            #plt.xlim(xmin = 0, xmax = 12)
            plt.scatter(dis1, dis2, s=5, edgecolors='k') 
            if i < 2:
                regr = linear_model.LinearRegression()    
                regr.fit(np.array(dis1).reshape(-1,1), np.array(dis2))
                Y = regr.predict(np.array(dis1).reshape(-1,1))
                plt.legend(['%.3f + %.3f*dis ' % (regr.intercept_ , regr.coef_[0])])
                plt.ylim(ymin = 0, ymax = 2)
                plt.plot(dis1, Y, color='blue', linewidth=2)
            else:
                plt.ylim(ymin = 0, ymax = 2)
        plt.show()     
    
    def plot_claster(self, words):               
        center = self.aver_vector(words)
        dist1, dist2 = [], []
        for w in words:
            dist1.append(self.dist_len(center, self.vec(w)).item())
            dist2.append(self.dist_cos(center, self.vec(w)).item())
        plt.scatter(dist1, dist2, s=5, edgecolors='k') 
        for i,w in enumerate(words):
            plt.text(dist1[i],dist2[i], w)
    
    def  bin_classifier1(self,  doc, s1, s2):        
        v1 = self.sent_to_vec(s1)                         # правильное предложение
        v2 = self.sent_to_vec(s2)                         # неправильное предложение        
        u  = self.sent_to_vec(doc)
                  
        return int(  self.dist(v1,u) <  self.dist(v2,u) )
        
    def  bin_classifier2(self,  doc, s1, s2):        
        v1 = self.sent_to_vec(s1)                         # правильное предложение
        v2 = self.sent_to_vec(s2)                         # неправильное предложение
        
        ss = torch.zeros( len(doc), self.vectors.shape[1] )
        for i,s in enumerate(doc):
            ss[i] = self.sent_to_vec(s)
        ss /= ss.norm(dim=1)
        v1 /= v1.norm(dim=1)
        v2 /= v2.norm(dim=1)
        
        u1, u2 = ss @ v1, ss @ v2
          
        return int(  u1.max() >  u2.max() )        